/**********************************************************************
		NOTEMIDI - module for serial port MIDI devices

		(C) 2001 by Michael Minn - michael@michaelminn.com

	Kernel source MUST be loaded on the system in order to compile this
	module.  Because modules cannot be dynamically loaded into the Linux
	Sound System (OSS), this module uses functions from the kernel source
	tree that enables this module to be listed on and accessed through
	/dev/sequencer.

	printk messages are written to /var/log/messages

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

***************************************************************************/

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/soundcard.h>

/* These include files are part of the kernel source tree */
#include <drivers/sound/sound_config.h>
#include <drivers/sound/dev_table.h>

#define NOTEMIDI_SERIAL_PORT_DEVICE_NAME "/dev/ttyS0"
#define NOTEMIDI_READ_TIMER_INTERVAL 5

/*************************************************************************
*	Because the OSS API doesn't implement any way of storing private
*	data and only references devices by device number, context information
*	must be stored in this global array indexed by device number
**************************************************************************/

static struct notemidi_device
{
	int is_open;
	int is_notemidi;
	int ms124w_protocol;
	struct midi_operations oss_operations;
} notemidi_devices[MAX_MIDI_DEV];

static struct file *notemidi_serial_port_file = NULL;
static struct tty_struct *notemidi_serial_port_tty = NULL;


/**********************************************************************************
*   Serial Port write/read
*
*	For info on the TTY serial port drivers, see
*	/usr/include/linux/tty_driver.h
*	/usr/src/linux/drivers/char/serial.c
*
*	Because serial port handling is extremely complex, NOTEMIDI uses the kernel
*	serial port functions.  Serial port input calls the line discipline functions
*	directly. This is not portable.  Reads from the line discipline do not block, so
*	schedule_timeout is used to poll the read function.
*
*	For MS-124 write protocol - one Burst Byte preceeds each data byte
*	lower 4 bits always 8
*	upper 4 bits correspond to output port
*		0x18 = port 1
*		0x28 = port 2
*		0x48 = port 3
*		0x88 = port 4
*		0xf8 = all four ports simultaneously
**********************************************************************************/

static struct tty_struct *notemidi_open_serial_port(char *serial_port_device_name)
{
	struct tty_struct *tty = NULL;
	struct termios old_termios;

	if (IS_ERR(notemidi_serial_port_file = filp_open(serial_port_device_name, O_WRONLY, 0)))
		printk(KERN_INFO "NOTEMIDI: Failure %d opening serial port %s\n", (int) PTR_ERR(notemidi_serial_port_file), serial_port_device_name);

	else if ((tty = (struct tty_struct *)notemidi_serial_port_file->private_data) == NULL)
		printk(KERN_INFO "NOTEMIDI: NULL file private data TTY pointer\n");

	else if (tty->driver.set_termios == NULL)
		printk(KERN_INFO "NOTEMIDI: NULL file set_termios function\n");

	else if (tty->termios == NULL)
		printk(KERN_INFO "NOTEMIDI: NULL file termios structure\n");

	else if (tty->ldisc.flush_buffer == NULL)
		printk(KERN_ERR "NOTEMIDI: NULL TTY ldisc.flush_buffer() function\n");

	else if (tty->driver.write == NULL)
		printk(KERN_ERR "NOTEMIDI: NULL TTY driver write function\n");

	else if (tty->ldisc.read == NULL)
		printk(KERN_ERR "NOTEMIDI: NULL TTY ldisc.read() function\n");

	else
	{
		old_termios = *tty->termios;
		tty->termios->c_lflag = 0;
		tty->termios->c_lflag = NOFLSH;
		tty->termios->c_iflag = IGNBRK | IGNPAR;
		tty->termios->c_oflag = 0;
		tty->termios->c_cflag =  B38400 | CREAD | CSIZE | CS8 | CLOCAL | HUPCL;
		tty->termios->c_cc[VEOL] = 0; /* '\r'; */
		tty->termios->c_cc[VERASE] = 0;
		tty->termios->c_cc[VKILL] = 0;
		tty->termios->c_cc[VMIN] = 0;
		tty->termios->c_cc[VTIME] = 0;
		(*tty->driver.set_termios)(tty, &old_termios);
		return tty;
	}

	return 0;
}

static int notemidi_open (struct notemidi_device *device)
{
	printk(KERN_INFO "NOTEMIDI: open device /dev/midi%02d\n", device - notemidi_devices);

	if (device->is_open)
		return -EBUSY;

	else
	{
		/*	MOD_INC_USE_COUNT;	*/
		/* tty->ldisc.flush_buffer(tty); */
		++device->is_open;
	}

	return 0;
}

static int notemidi_close (struct notemidi_device *device)
{
	printk(KERN_INFO "NOTEMIDI: close device /dev/midi%02d\n", device - notemidi_devices);

	/*	MOD_DEC_USE_COUNT;	*/
	--device->is_open;
	return 0;
}

static int notemidi_write (struct notemidi_device *device, unsigned char data)
{
	int status = -EFAULT;
	char burst_byte = 0xf8;

	if (device->ms124w_protocol && ((status = notemidi_serial_port_tty->driver.write(notemidi_serial_port_tty, 0, &burst_byte, 1)) < 1))
		printk(KERN_ERR "NOTEMIDI: ms124w burst byte write failure: status %d\n", status);

	else if ((status = notemidi_serial_port_tty->driver.write(notemidi_serial_port_tty, 0, &data, 1)) < 1)
		printk(KERN_ERR "NOTEMIDI: write failure: status %d\n", status);

	return status;
}

static int notemidi_read (struct notemidi_device *device, unsigned char *data)
{
	int status = 0;

	/* if (signal_pending(current)) */
	while ((status = notemidi_serial_port_tty->ldisc.read(notemidi_serial_port_tty, notemidi_serial_port_file, data, 1)) == 0)
		schedule_timeout(jiffies + 5); /* an arbitrary value of clock ticks: 5 = 1/20 second */

	return status;
}


/****************************************************************************
*	/dev/midiXX file operations
*
*	It is possible to override the standard OSS file_operations functions
*	with module supplied file_operations.  There should be no conflicts
*	with interrupts from /dev/sequencer which calls the OSS midi_operations.
*
*****************************************************************************/


static int notemidi_file_open (struct inode *inode, struct file *file)
{
	if ((MINOR(inode->i_rdev) >> 4) >= MAX_MIDI_DEV)
		return -EINVAL;
	else
	{
		file->private_data = &notemidi_devices[MINOR(inode->i_rdev) >> 4];
		return notemidi_open(&notemidi_devices[MINOR(inode->i_rdev) >> 4]);
	}
}

static int notemidi_file_release (struct inode *inode, struct file *file)
{
	return notemidi_close((struct notemidi_device*)file->private_data);
}

static ssize_t notemidi_file_write (struct file *file, const char *data, size_t length, loff_t *offset)
{
	int scan;
	for (scan = 0; scan < length; ++scan)
	{
		int status;
		unsigned char buffer = 0;
		get_user(buffer, &data[scan]);
		if ((status = notemidi_write((struct notemidi_device*)file->private_data, buffer)) < 1)
			return status;
	}

	return length;
}


static ssize_t notemidi_file_read (struct file *file, char *data, size_t length, loff_t *offset)
{
	/* unsigned char buffer; */
	int status = notemidi_read ((struct notemidi_device*)file->private_data, data);
	/* put_user(buffer, data); */
	return status;
}

struct file_operations notemidi_file_operations =
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
    NULL, /* struct module *owner; */
#endif
	NULL, /* loff_t (*llseek) (struct file *, loff_t, int); */
	notemidi_file_read, /* ssize_t (*read) (struct file *, char *, size_t, loff_t *); */
	notemidi_file_write, /* ssize_t (*write) (struct file *, const char *, size_t, loff_t *); */
	NULL, /* int (*readdir) (struct file *, void *, filldir_t); */
	NULL, /* unsigned int (*poll) (struct file *, struct poll_table_struct *); */
	NULL, /* int (*ioctl) (struct inode *, struct file *, unsigned int, unsigned long); */
	NULL, /* int (*mmap) (struct file *, struct vm_area_struct *); */
	notemidi_file_open, /* int (*open) (struct inode *, struct file *); */
	NULL, /* int (*flush) (struct file *); */
	notemidi_file_release, /* int (*release) (struct inode *, struct file *); */
	NULL, /* int (*fsync) (struct file *, struct dentry *); */
	NULL, /* int (*fasync) (int, struct file *, int); */
	NULL, /* int (*lock) (struct file *, int, struct file_lock *); */
	NULL, /* ssize_t (*readv) (struct file *, const struct iovec *, unsigned long, loff_t *); */
	NULL, /* ssize_t (*writev) (struct file *, const struct iovec *, unsigned long, loff_t *); */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
	NULL, /* ssize_t (*writepage) (struct file *, struct page *, int, size_t, loff_t *, int); */
#endif
};


/**********************************************************************************
*
*	OSS OPERATIONS & REGISTRATION
*
*	In order for MIDI devices to be visible to the Linux Open Sound System (OSS)
*	they must be registered with a struct midi_operations.  However, the
*	interface is designed to facilitate programming of very low level devices
*	and is rather cumbersome. The operations functions can be overriden for
*	use with the raw midi files (/dev/midiXX), but read/write to /dev/sequencer
*	calls these functions directly, so they must be implemented to permit
*	compatibility with programs that use /dev/sequencer. Only the open, close
*	and write (outputc) midi_operations appear to be used by any OSS functions.
*	The text names listed by ioctl() of /dev/sequencer is also stored in this structure.
*
*	The OSS MIDI buffering mechanism (/usr/src/linux/drivers/sound/midibuf.c)
*	implements blocking reads with a sleep que and the serial port implements
*	non-blocking input. I attempted to trap the serial port interrupt but
*	there were unacceptable latency problems.  I also attempted to use a kernel timer
*	to poll the serial port, but conflicts with the /dev/sequencer timers resulted
*	in the dreaded fatal kernel interrupt crash ("Aiee, killing interrupt handler").
*	Therefore, MIDI input is available to the /dev/midiXX ports, but not /dev/sequencer.
*
*	For info on /dev/sequencer, see /usr/src/linux/drivers/char/serial.c
*
***********************************************************************************/

static int notemidi_oss_open (int device, int mode, void (*input_callback)(int dev, unsigned char data), void (*output_callback)(int dev))
{
	if (device >= MAX_MIDI_DEV)
		return -EINVAL;
	else
		return notemidi_open(&notemidi_devices[device]);
}

static void notemidi_oss_close (int device)
{
	if (device < MAX_MIDI_DEV)
		notemidi_close(&notemidi_devices[device]);
}

static int notemidi_oss_write (int device, unsigned char data)
{
	if (device >= MAX_MIDI_DEV)
		return -EINVAL;
	else
		return notemidi_write(&notemidi_devices[device], data);

}


/************************ Module entry/exit points ********************************/

int init_module(void)
{
	int notemidi_minor_number = 0;
	int ms124w_minor_number = 0;
	printk(KERN_INFO "NOTEMIDI: Loading Linux Serial Port MIDI interface module v2001.06.01c\n");

	if ((notemidi_serial_port_tty = notemidi_open_serial_port(NOTEMIDI_SERIAL_PORT_DEVICE_NAME)) == NULL)
		return -1;

	else if ((notemidi_minor_number = register_sound_midi(&notemidi_file_operations, -1)) < 0)
		printk(KERN_ERR "NOTEMIDI: register_sound_midi() failed\n");

	else if ((ms124w_minor_number = register_sound_midi(&notemidi_file_operations, -1)) < 0)
		printk(KERN_ERR "NOTEMIDI: MS-124W register_sound_midi() failed\n");

	else
	{
		printk(KERN_ERR "NOTEMIDI: /dev/midi%02d and (ms124w) /dev/midi%02d\n", notemidi_minor_number >> 4, ms124w_minor_number >> 4);
		memset(&notemidi_devices, 0, sizeof(notemidi_devices));

		notemidi_devices[notemidi_minor_number >> 4].is_notemidi = 1;
		strcpy(notemidi_devices[notemidi_minor_number >> 4].oss_operations.info.name, "NOTEMIDI");
		notemidi_devices[notemidi_minor_number >> 4].oss_operations.open = notemidi_oss_open;
		notemidi_devices[notemidi_minor_number >> 4].oss_operations.outputc = notemidi_oss_write;
		notemidi_devices[notemidi_minor_number >> 4].oss_operations.close = notemidi_oss_close;
		midi_devs[notemidi_minor_number >> 4] = &notemidi_devices[notemidi_minor_number >> 4].oss_operations;

		notemidi_devices[ms124w_minor_number >> 4].is_notemidi = 1;
		notemidi_devices[ms124w_minor_number >> 4].ms124w_protocol = 1;
		strcpy(notemidi_devices[ms124w_minor_number >> 4].oss_operations.info.name, "MS-124W");
		notemidi_devices[ms124w_minor_number >> 4].oss_operations.open = notemidi_oss_open;
		notemidi_devices[ms124w_minor_number >> 4].oss_operations.outputc = notemidi_oss_write;
		notemidi_devices[ms124w_minor_number >> 4].oss_operations.close = notemidi_oss_close;
		midi_devs[ms124w_minor_number >> 4] = &notemidi_devices[ms124w_minor_number >> 4].oss_operations;

		if((ms124w_minor_number >> 4) >= num_midis)
			num_midis = (ms124w_minor_number >> 4) + 1;
	}

	return 0;
}

int cleanup_module(void)
{
	int scan;
	printk(KERN_INFO "NOTEMIDI: Unloading module\n");

	if (!IS_ERR(notemidi_serial_port_file))
		filp_close(notemidi_serial_port_file, 0);

	for (scan = MAX_MIDI_DEV - 1; scan >= 0; --scan)
		if (notemidi_devices[scan].is_notemidi)
		{
			notemidi_devices[scan].is_notemidi = 0;
			sound_unload_mididev(scan);
			if (num_midis == (scan + 1))
				--num_midis;
		}

	return 0;
}
